import random
from textual.widgets import Label, Button
from rich.style import Style
from rich.console import RenderableType
from rich.segment import Segments, Segment
# > Typing
from typing import Optional, Union, Callable, Awaitable
from inspect import iscoroutinefunction

# ! Fill Label Class
class FillLabel(Label):
    DEFAULT_CSS = """
    FillLabel {
        height: 1fr;
        width: 1fr;
    }
    """
    def __gen_random_color(self) -> None:
        return f"#{hex(random.randint(0,16777215))[2:]}"
    
    def __gen_segments(self) -> Segments:
        segments = []
        for i in range(self.size[0]*self.size[1]):
            s = Style.parse(self.__gen_random_color()) if self.__rainbow else self.__style
            segments.append(Segment(self.__chr, s))
        return Segments(segments)
    
    def __init__(
        self,
        char: str="-",
        style: Optional[Style]=None,
        rainbow: bool=False,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.__chr = char[:1]
        self.__rainbow = rainbow
        self.__style = style
        self.update(self.__gen_segments())
    
    async def on_resize(self) -> None:
        self.update(self.__gen_segments())

# ! Clickable Label Class
class ClickableLabel(Label, Button, can_focus=True):
    DEFAULT_CSS = """
    ClickableLabel {
        width: auto;
        min-width: 1;
        height: auto;
        min-height: 1;
        color: $text;
        text-style: bold;
        text-align: center;
        content-align: center middle;
    }
    ClickableLabel:focus {
        text-style: bold reverse;
    }
    ClickableLabel:hover {
        color: $text;
    }
    ClickableLabel.-active {
        tint: $background 30%;
    }
    """
    def __init__(
        self,
        renderable: RenderableType="",
        callback: Union[Callable[[], None], Callable[[], Awaitable[None]]]=lambda: None,
        *,
        expand: bool=False,
        shrink: bool=False,
        markup: bool=True,
        name: Optional[str]=None,
        id: Optional[str]=None,
        classes: Optional[str]=None,
        disabled: bool=False
    ) -> None:
        super().__init__(
            renderable,
            expand=expand,
            shrink=shrink,
            markup=markup,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled
        )
        self.__callback = callback
    
    @property
    def callback_awaitable(self) -> bool:
        return iscoroutinefunction(self.__callback)
    
    async def _on_click(self, event) -> None:
        await super()._on_click(event)
        if self.callback_awaitable:
            await self.__callback()
        else:
            self.__callback()



import os
import glob
import properties
from typing import Dict, List, Tuple, Optional
# > Local Imports
from .functions import formater
from .exceptions import LanguageNotExistError, LanguageNotLoadedError

# ! Child Class
class Language:
    """The class reflects the file with the translation."""
    @staticmethod
    def __load_file(filepath: str) -> Dict[str, str]:
        with open(filepath, 'r', errors='ignore', encoding='utf-8') as file:
            return properties.load(file)
    
    def __get_metadata(self) -> Tuple[str, str, Optional[str], Dict[str, str]]:
        data = self.__load_file(self.__name)
        return \
            data.get("language.metadata.title", f"<LTNF:{self.__mark}>"), \
            data.get("language.metadata.author", "<unknown>"), \
            data.get("language.metadata.author.url", None), \
            {
                "from": data.get("language.metadata.words.from", "from")
            }
    
    # ? Initialization
    def __init__(self, language_filepath: str) -> None:
        """The class reflects the file with the translation.
        
        Args:
            language_filepath (str): The path to the file with the translation.
        
        Raises:
            FileNotFoundError: It is called if the file with the translation does not exist.
        """
        self.__name = os.path.abspath(language_filepath)
        self.__data, self.__loaded = None, False
        # * Checking
        if not os.path.isfile(self.__name):
            raise FileNotFoundError
        # * Metadata Loading
        self.__mark = os.path.splitext(os.path.basename(self.__name))[0].lower()
        self.__title, self.__author, self.__author_url, self.__words = self.__get_metadata()
    
    # ? Magic Methods
    def __str__(self) -> str:
        form = formater(name=self.__name, title=self.__title, mark=self.__mark, loaded=self.__loaded)
        return f"{self.__class__.__name__}({form})"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    # ? Propertyes
    @property
    def name(self) -> str:
        """The path to the file with the translation.
        
        Returns:
            The full path to the file.
        """
        return self.__name
    
    @property
    def mark(self) -> str:
        """The name of the file without an extension.
        
        Returns:
            The name of the file without the extension.
        """
        return self.__mark
    
    @property
    def title(self) -> str:
        """The name of the translation language taken from the file.
        
        Returns:
            The purpose of the language.
        """
        return self.__title
    
    @property
    def author(self) -> str:
        """The author of the translation taken from the translation file.
        
        Returns:
            The author's nickname.
        """
        return self.__author
    
    @property
    def author_url(self) -> Optional[str]:
        """The link to the author of the translation is taken from the translation file.
        
        Returns:
            Link to the author.
        """
        return self.__author_url
    
    @property
    def words(self) -> Dict[str, str]:
        """Special values that you need to have even if the language is not loaded.
        
        Returns:
            Dictionary of special meanings.
        """
        return self.__words
    
    @property
    def loaded(self) -> bool:
        """If `True`, then the file with the translation is fully loaded in memory.
        
        Returns:
            If loaded, then `True`.
        """
        return self.__loaded
    
    # ? Main Methods
    def load(self) -> None:
        """Full load of the file with the translation in memory."""
        self.__data, self.__loaded = self.__load_file(self.__name), True
    
    def unload(self) -> None:
        """Unload a file with a translation from memory."""
        self.__data, self.__loaded = None, False
    
    def get(self, key: str, default: Optional[str]=None) -> Optional[str]:
        """Getting a line feed.
        
        Args:
            key (str): The key to the variable with the translation, which is registered in the file with the translation.
            default (Optional[str], optional): The default value. Defaults to None.
        
        Raises:
            LanguageNotLoadedError: It is called if the translation file is not fully loaded into memory.
        
        Returns:
            Optional[str]: The translated string.
        """
        if self.loaded:
            return self.__data.get(key, default)
        raise LanguageNotLoadedError()

# ! Main Class
class LanguageLoader:
    """The loader of files with translation."""
    # ? Main Methods
    def __search_langs(self, dlm: str, mlm: str) -> Tuple[Language, Optional[Language]]:
        dl, ml = None, None
        for lang in self.langs:
            if lang.mark == dlm:
                dl = lang
            if lang.mark == mlm:
                ml = lang
        if dl is None:
            raise LanguageNotExistError([dl])
        dl.load()
        if ml is not None:
            ml.load()
        return dl, ml
    
    # ? Initialization
    def __init__(
        self,
        langs_dirpath: str,
        main_lang_mark: str,
        default_lang_mark: str="en-eng",
    ) -> None:
        """The loader of files with translation.
        
        Args:
            langs_dirpath (str): The path to the folder with the translation files.
            main_lang_mark (str): The name of the file with the main translation without the extension.
            default_lang_mark (str, optional): The name of the file with the default translation without an extension. Defaults to "en-eng".
        
        Raises:
            FileNotFoundError: Called if the file with the default translation without the extension could not be found.
        """
        self.__name = os.path.abspath(langs_dirpath)
        self.__mlm = main_lang_mark.lower()
        self.__dlm = default_lang_mark.lower()
        self.__alangs: List[LanguageLoader] = []
        # * Checking
        if not os.path.isdir(self.__name):
            raise FileNotFoundError
        # * Searching and loading languages
        self.__langs: List[Language] = [Language(lp) for lp in glob.glob(os.path.join(self.__name, "??-???.properties"))]
        self.__dlang, self.__mlang = self.__search_langs(self.__dlm, self.__mlm)
    
    # ? Magic Methods
    def __str__(self) -> str:
        form = formater(
            name=self.__name,
            default_lang_mark=self.__dlm,
            main_lang_mark=self.__mlm,
            langs=self.__langs,
            alangs=self.__alangs
        )
        return f"{self.__class__.__name__}({form})"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    # ? Propertys
    @property
    def name(self) -> str:
        """The path to the folder with the translation files.
        
        Returns:
            The full path to the file.
        """
        return self.__name
    
    @property
    def main_lang_mark(self) -> str:
        """The name of the file with the main translation without the extension.
        
        Returns:
            The file name without an extension.
        """
        return self.__mlm
    
    @property
    def default_lang_mark(self) -> str:
        """The name of the file with the default translation without an extension.
        
        Returns:
            The file name without an extension.
        """
        return self.__dlm
    
    @property
    def langs(self) -> List[Language]:
        """A list with images of the `Language` class, reflecting the files found in the folder with the translation files.
        
        Returns:
            A list with translations.
        """
        return self.__langs
    
    @property
    def default_lang(self) -> Language:
        """An image of the `Language` class reflecting the uploaded file with the default translation.
        
        Returns:
            The translation.
        """
        return self.__dlang
    
    @property
    def main_lang(self) -> Optional[Language]:
        """An image of the `Language` class reflecting the uploaded file with the main translation.
        
        Returns:
            The translation or `None`.
        """
        return self.__mlang
    
    @property
    def alangs(self):
        """Additional languages, for example, translation of plugins.
        
        Returns:
            A list with an additional translation.
        """
        return self.__alangs
    
    # ? Public Methods
    def get(self, key: str) -> str:
        """First, it tries to get the translation from the file with the main translation, if it failed, it tries to get it from the file with the default translation, if it failed again, it tries to get them from additional languages (`self.alangs`). If in the end None is still output, then returns the string `"<LTNF>"`, that is, the `Language Text Not Found`.
        
        Args:
            key (str): The key to the variable with the translation, which is registered in the file with the translation.
        
        Returns:
            str: The translated string.
        """
        if self.__mlang is None:
            data = self.__dlang.get(key)
        data = self.__mlang.get(key, self.__dlang.get(key))
        if data is None:
            for alang in self.__alangs:
                if (data:=alang.get(key)) is not None:
                    break
        return data if data is not None else "<LTNF>"
    
    def merge(self, ll) -> None:
        """Adding additional languages (`self.alangs`).
        
        Args:
            ll (LanguageLoader): The image of the `LanguageLoader` class.
        """
        assert isinstance(ll, LanguageLoader)
        self.__alangs.append(ll)


from pydantic import BaseModel
from textual.screen import Screen
from textual.binding import Binding
# > Typing
from typing import Optional, Generator, Type, Callable, List, Any
# > Local Import's
from ..seaplayer import SeaPlayer
from ..codeсbase import CodecBase
from .pluginloader import PluginLoader

# ! Plugin Info Class
class PluginInfo(BaseModel):
    name: str
    name_id: str
    version: str
    author: str
    description: Optional[str]=None
    url: Optional[str]=None

# ! Plugin Base Class
class PluginBase:
    """The base class of the plugin."""
    app: SeaPlayer
    """A link to the image of the `SeaPlayer` class."""
    pl: PluginLoader
    """A link to the image of the `PluginLoader` class."""
    info: PluginInfo
    """Contains all variables from `info.json`."""
    
    def __init_repr__(self) -> str:
        """The string that is output to the logs after initialization of the class.
        
        Returns:
            str: A string indicating the initialization of the class.
        """
        ...
    def __init__(self, app: SeaPlayer, pl: PluginLoader, info: PluginInfo) -> None:
        """The base class of the plugin.
        
        Args:
            app (SeaPlayer): A link to the image of the `SeaPlayer` class.
            pl (PluginLoader): A link to the image of the `PluginLoader` class.
            info (PluginInfo): Contains all variables from `info.json`.
        """
        ...
    # ! App Specific Methods
    def install_screen(self, name: str, screen: Screen) -> None:
        """Adding a new `Screen`.
        
        Args:
            name (str): The name by which it will be accessed by SeaPlayer.
            screen (Screen): The image of the `Screen` class.
        """
        ...
    def add_codecs(self, *codecs: Type[CodecBase]) -> None:
        """Adding codecs."""
        ...
    def add_value_handlers(self, *handlers: Callable[[str], List[str]]) -> None:
        """Adding handlers for user-entered values."""
        ...
    # ! App Specific On Methods
    def on_bindings(self) -> Generator[Binding, Any, None]:
        """A generator for binding information classes about keys.
        
        Yields:
            Generator[Binding, Any, None]: A class with information about key bindings.
        """
        ...
    # ! App On Methods
    def on_init(self) -> None:
        """A function called during the initialization of the SeaPlayer."""
        ...
    def on_run(self) -> None:
        """A function called before launching SeaPlayer."""
        ...
    async def on_compose(self) -> None:
        """A function called after initializing the SeaPlayer widgets."""
        ...
    async def on_ready(self) -> None:
        """The function starts immediately after the SeaPlayer is fully launched."""
        ...
    async def on_quit(self) -> None:
        """The function starts after sending a signal to close the SeaPlayer."""
        ...



import asyncio
from rich.text import TextType
from textual.widgets import Button
# > Typing
from typing import Optional, Callable, Awaitable

# ! Clikable Button Class
class ClikableButton(Button):
    """The widget responds to clicks."""
    def __init__(
        self,
        label: str,
        on_click: Callable[[], Awaitable[None]],
        *args,
        **kwargs
    ) -> None:
        """A widget that calls the `on_click` function after clicking on it.
        
        Args:
            label (str): The text inside the widget.
            on_click (Callable[[], Awaitable[None]]): The function that is called after clicking on the widget.
        """
        super().__init__(label, *args, **kwargs)
        self.__on_click_method = on_click
    
    async def on_click(self, *args, **kwargs) -> None:
        await self.__on_click_method()

# ! Wait Button Click
class WaitButton(Button):
    """A widget with the function of waiting for a click."""
    def __init__(
        self,
        label: Optional[TextType]=None,
        queue_max_size: int=1,
        *args,
        **kwargs
    ) -> None:
        """_summary_
        
        Args:
            label (Optional[TextType], optional): _description_. Defaults to None.
            queue_max_size (int, optional): _description_. Defaults to 1.
        """
        super().__init__(label, *args, **kwargs)
        self.__wait_queue: asyncio.Queue[bool] = asyncio.Queue(queue_max_size)
    
    async def on_click(self, *args, **kwargs) -> None:
        await self.__wait_queue.put(True)
    
    async def wait_click(self) -> bool:
        """A function that includes waiting for a click.
        
        Returns:
            bool: Always returns `True`.
        """
        return await self.__wait_queue.get()


from textual.widgets import Input
# > Typing
from typing_extensions import deprecated
from typing import Optional, Tuple, Callable, Awaitable, Any

# ! InputFiles functions
async def conv(value: str) -> Tuple[bool, Optional[Any]]: return True, value
async def submit(input: Input, value: Any) -> None: ...
def update_placeholder() -> Optional[str]: ...

# ! InputField class
@deprecated("It is planned to remove this class.")
class InputField(Input):
    def __init__(
        self,
        conv: Callable[[str], Awaitable[Tuple[bool, Optional[Any]]]]=conv,
        submit: Callable[[Input, Any], Awaitable[None]]=submit,
        update_placeholder: Callable[[], Optional[str]]=update_placeholder,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.__conv = conv
        self.__submit = submit
        self.__update_placeholder = update_placeholder
        if (placeholder:=self.__update_placeholder()) is not None:
            self.placeholder = placeholder
    
    async def action_submit(self):
        value = self.value
        self.value = ""
        if value.replace(" ", "") != "":
            ok, c_value = await self.__conv(value)
            if ok:
                await self.__submit(self, c_value)
        if (placeholder:=self.__update_placeholder()) is not None:
            self.placeholder = placeholder



import os
import validators
from typing import Optional

# ! Functions
def formater(**kwargs) -> str:
    return ", ".join([f"{key}={repr(value)}" for key, value in kwargs.items()])

# ! Codecs Base
class CodecBase:
    """The base class of the codec."""
    # * Codec Info
    codec_name: str="None"
    """The name of the codec (abbreviation)."""
    codec_priority: float=0.0
    """Sorting priority (the lower the value, the earlier it will be processed, checked for compatibility and initialized)"""
    
    # * Info
    name: str
    """File path (optional file path)."""
    hidden_name: bool=False
    """Hide the `name` when displayed in the SeaPlayer."""
    duration: float
    """The duration of the sound in seconds."""
    channels: int
    """The number of channels in the file."""
    samplerate: int
    """The sampling rate in hertz (Hz)."""
    bitrate: int
    """The number of bits per second of playback (determines the sound quality)."""
    
    # * Sound Info
    title: Optional[str]
    """The `title` of the sound is taken from the metadata."""
    artist: Optional[str]
    """The `artist` is taken from the metadata."""
    album: Optional[str]
    """The `album title` is taken from the metadata."""
    icon_data: Optional[bytes]
    """A `poster` in bytes format is obtained from metadata."""
    
    # * Playback Info
    playing: bool
    """The playback status (`True if playing`)."""
    paused: bool
    """The playback status (`True if paused`)."""
    
    # ! Initializing Functions
    def __init__(self, path: str, **kwargs) -> None:
        """The base class of the codec.
        
        Args:
            path (str): File path (optional file path).
        """
        ...
    #async def __aio_init__(self, path: str, **kwargs) -> None: ...
    def __repr__(self):
        return \
            "{0}({1})".format(
                self.__class__.__name__,
                formater(
                    name=self.name,
                    duration=self.duration,
                    channels=self.channels,
                    samplerate=self.samplerate,
                    bitrate=self.bitrate,
                    title=self.title,
                    artist=self.artist,
                    album=self.album
                )
            )
    
    def __namerepr__(self) -> str:
        """The name that is displayed for the place `self.name`.
        
        Returns:
            str: Value `self.name` the displayed in SeaPlayer.
        """
        if self.hidden_name:
            return ""
        if self.name is not None:
            return self.name
        return f'<memory>'
    
    def __sha1__(self, buffer_size: int) -> str:
        """Calculating the hash of the file.
        
        Args:
            buffer_size (int): The size of the temporary buffer.
        
        Returns:
            str: SHA256 in string format.
        """
        ...
    
    async def __aio_sha1__(self, buffer_size: int) -> str:
        """Calculating the hash of the file.
        
        Args:
            buffer_size (int): The size of the temporary buffer.
        
        Returns:
            str: SHA256 in string format.
        """
        ...
    
    # ! Testing Functions
    @staticmethod
    def is_this_codec(path: str) -> bool:
        """Compatibility check.
        
        Args:
            path (str): File path (optional file path).
        
        Returns:
            bool: True if compatible.
        """
        return False
    
    @staticmethod
    async def aio_is_this_codec(path: str) -> bool:
        """Compatibility check.
        
        Args:
            path (str): File path (optional file path).
        
        Returns:
            bool: True if compatible.
        """
        return False
    
    # ! Playback Functions
    def play(self) -> None:
        """Start playing the sound."""
        ...
    
    def stop(self) -> None:
        """Stop playing the sound."""
        ...
    
    def pause(self) -> None:
        """Put it on pause."""
        ...
    
    def unpause(self) -> None:
        """Take it off the pause."""
        ...
    
    def get_volume(self) -> float:
        """Getting the current volume as a percentage.
        
        Returns:
            float: Volume percentage (0.01 == 1%).
        """
        return 1.0
    
    def set_volume(self, value: float) -> None:
        """Setting the volume value as a percentage.
        
        Args:
            value (float): Volume percentage (0.01 == 1%).
        """
        ...
    
    def get_pos(self) -> float:
        """Getting the audio playback position in seconds.
        
        Returns:
            float: The position of the audio playback in seconds.
        """
        return 0.0
    
    def set_pos(self, value: float) -> None:
        """Setting the audio playback position in seconds.
        
        Args:
            value (float): The position of the audio playback in seconds.
        """
        ...



from textual.widget import Widget
from textual.widgets import Label
from textual.containers import Container
# > Typing
from typing import Optional
# > Local Imports
from .Labels import FillLabel

# ! Main Class
class PopUp(Container):
    DEFAULT_CSS = """
    PopUp {
        layer: popup;
        background: $background;
        text-align: center;
        align-vertical: middle;
        align-horizontal: center;
        content-align: center middle;
        width: auto;
        height: auto;
        padding: 1 2 1 2;
    }
    """
    pass

# ! Other Clases
class PopUpWindow(PopUp):
    DEFAULT_CSS = """
    PopUpWindow {
        layer: popup;
        align-vertical: middle;
        align-horizontal: center;
        content-align: center middle;
        text-align: center;
    }
    PopUpWindow FillLabel {
        width: auto;
        height: 1;
    }
    """
    def __init__(
        self,
        *children: Widget,
        title: Optional[str]=None,
        name: Optional[str]=None,
        id: Optional[str]=None,
        classes: Optional[str]=None,
        disabled: bool=False
    ) -> None:
        self.__title = title or self.__class__.__name__
        self.__title_label = Label(self.__title)
        super().__init__(
            self.__title_label,
            FillLabel('─'),
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled
        )
    
    # ! Propertys
    @property
    def title(self) -> str:
        return self.__title
    
    @title.setter
    def title(self, value: str) -> str:
        self.__title = str(value)
        self.__title_label.update(self.__title)


from textual.containers import ScrollableContainer, Container
# > Typing
from typing import Optional, Union

# ! Child Class
class ConfigurateListItem(Container):
    DEFAULT_CSS = """
    ConfigurateListItem {
        border: solid dodgerblue;
        background: #0000;
        border-title-color: #aaaaaa;
        border-subtitle-color: #6b6b6b;
        height: auto;
    }
    """
    
    def __init__(
        self,
        *children,
        title: str="",
        desc: str="",
        width: Union[int, str]="1fr",
        height: Union[int, str]="1fr",
        **kwargs
    ) -> None:
        super().__init__(*children, **kwargs)
        self.border_title = title
        self.border_subtitle = desc
        self.styles.width = width
        self.styles.height = height
    
    async def updating(self, title: Optional[str]="", desc: Optional[str]="") -> None:
        if title is not None: self.border_title = title
        if desc is not None: self.border_subtitle = desc

# ! Main Class
class ConfigurateList(ScrollableContainer):
    DEFAULT_CSS = """
    ConfigurateList {
        border: solid cadetblue;
        height: 1fr;
        width: 1fr;
    }
    """
    
    def __init__(self, *children, **kwargs) -> None:
        super().__init__(*children, **kwargs)



from .Log import LogMenu
from .Labels import FillLabel, ClickableLabel
from .Input import InputField
from .Buttons import ClikableButton, WaitButton
from .Notification import Nofy, CallNofy
from .ProgressBar import IndeterminateProgress
from .DataOptions import DataOption, DataOptionList
from .Image import AsyncImageLabel, StandartImageLabel
from .PlayList import PlayListView, PlayListViewItem
from .Configurate import ConfigurateListItem, ConfigurateList
from .Radio import DataRadioSet, DataRadioButton
from .PopUp import PopUp, PopUpWindow
from .Rheostat import Rheostat



import properties
from pathlib import Path
from typing import Dict, Any, Optional, TypeVar, Union, Literal

# ! Types
T = TypeVar("T")

# ! Vars
DEFAULT_CONFIG_DATA: Dict[str, Any] = {
    "main.lang": "en-eng",
    "sound.sound_font_path": None,
    "sound.output_sound_device_id": None,
    "image.image_update_method": "sync",
    "image.image_resample_method": "bilinear",
    "playback.rewind_count_seconds": 5,
    "playback.volume_change_percent": 0.05,
    "playback.max_volume_percent": 2.0,
    "playlist.recursive_search": False,
    "keys.quit": "q,й",
    "keys.rewind_forward": "*",
    "keys.rewind_back": "/",
    "keys.volume_up": "+",
    "keys.volume_down": "-",
    "debag.logging": False
}
"""Default configuration values."""

# ! Main Class
class SeaPlayerConfig:
    """The main configuration class of the SeaPlayer."""
    @staticmethod
    def dump(filepath: Path, data: Dict[str, Any]) -> None:
        with open(filepath, "w", encoding="utf-8", errors="ignore") as file:
            properties.dump(data, file)
    
    @staticmethod
    def load(filepath: Path, default: Dict[str, Any]) -> Dict[str, Any]:
        with open(filepath, "r", encoding="utf-8", errors="ignore") as file:
            try:
                return properties.load(file)
            except:
                pass
        with open(filepath, "w", encoding="utf-8", errors="ignore") as file:
            properties.dump(default, file)
        return default
    
    def refresh(self) -> None:
        """Overwriting configurations to a file."""
        self.dump(self.filepath, self.config)
    
    def __init__(
        self,
        filepath: str,
        *,
        default_data: Dict[str, Any]=DEFAULT_CONFIG_DATA
    ) -> None:
        """The main configuration class of the SeaPlayer.
        
        Args:
            filepath (str): The path to the configuration file.
            default_data (Dict[str, Any], optional): Default configuration values. Defaults to DEFAULT_CONFIG_DATA.
        """
        self.filepath = Path(filepath)
        self.default_data = default_data
        if self.filepath.exists():
            self.config = self.load(self.filepath, self.default_data)
            config_temp = self.default_data.copy()
            config_temp.update(self.config)
            self.config = config_temp.copy()
            del config_temp
        else:
            self.config = default_data.copy()
        self.refresh()
    
    def get(self, key: str, default: T=None) -> Union[Any, T]:
        return self.config.get(key, default)
    
    def set(self, key: str, value: Any) -> None:
        self.config[key] = value
        self.refresh()
    
    # ! Main
    @property
    def lang(self) -> Union[Literal['en-eng'], str]:
        """The current language.
        
        Returns:
            The file name without an extension.
        """
        return self.get("main.lang")
    @lang.setter
    def lang(self, value: Union[Literal['en-eng'], str]) -> None: self.set("main.lang", value)
    
    # ! Sound
    @property
    def sound_font_path(self) -> Optional[str]:
        """The path to the file with the audio font.
        
        Returns:
            The full path to the file.
        """
        return self.get("sound.sound_font_path")
    @sound_font_path.setter
    def sound_font_path(self, value: Optional[str]): self.set("sound.sound_font_path", value)
    
    @property
    def output_sound_device_id(self) -> Optional[int]:
        """ID of the audio output device.
        
        Returns:
            The device ID.
        """
        return self.get("sound.output_sound_device_id")
    @output_sound_device_id.setter
    def output_sound_device_id(self, value: Optional[int]): self.set("sound.output_sound_device_id", value)
    
    # ! Image
    @property
    def image_update_method(self) -> Literal["sync", "async"]: return self.get("image.image_update_method")
    @image_update_method.setter
    def image_update_method(self, value: Literal["sync", "async"]): self.set("image.image_update_method", value)
    
    @property
    def image_resample_method(self) -> Literal["nearest", "bilinear", "bicubic", "lanczos", "hamming", "box"]:
        """The image resampling method.
        
        Returns:
            The image resampling method.
        """
        return self.get("image.image_resample_method")
    @image_resample_method.setter
    def image_resample_method(self, value: Literal["nearest", "bilinear", "bicubic", "lanczos", "hamming", "box"]):
        self.set("image.image_resample_method", value)
    
    # ! Playback
    @property
    def volume_change_percent(self) -> float:
        """Percentage by which the volume changes when the special keys are pressed.
        
        Returns:
            Value 0.01 == 1%.
        """
        return self.get("playback.volume_change_percent")
    @volume_change_percent.setter
    def volume_change_percent(self, value: float): self.set("playback.volume_change_percent", value)
    
    @property
    def rewind_count_seconds(self) -> int:
        """The value of the seconds by which the current sound will be rewound.
        
        Returns:
            Rewind in seconds.
        """
        return self.get("playback.rewind_count_seconds")
    @rewind_count_seconds.setter
    def rewind_count_seconds(self, value: int): self.set("playback.rewind_count_seconds", value)
    
    @property
    def max_volume_percent(self) -> float:
        """Maximum volume value.
        
        Returns:
            Value 0.01 == 1%.
        """
        return self.get("playback.max_volume_percent")
    @max_volume_percent.setter
    def max_volume_percent(self, value: float): self.set("playback.max_volume_percent", value)
    
    # ! Playlist
    @property
    def recursive_search(self) -> bool:
        """Recursive file search.
        
        Returns:
            On or off.
        """
        return self.get("playlist.recursive_search")
    @recursive_search.setter
    def recursive_search(self, value: bool): self.set("playlist.recursive_search", value)
    
    # ! Keys
    @property
    def key_quit(self) -> str:
        """The key to exit the SeaPlayer.
        
        Returns:
            The key char(s).
        """
        return self.get("keys.quit")
    @key_quit.setter
    def key_quit(self, value: str): self.set("keys.quit", value)
    
    @property
    def key_rewind_forward(self) -> str:
        """The rewind forward key.
        
        Returns:
            The key char(s).
        """
        return self.get("keys.rewind_forward")
    @key_rewind_forward.setter
    def key_rewind_forward(self, value: str): self.set("keys.rewind_forward", value)
    
    @property
    def key_rewind_back(self) -> str:
        """The rewind back key.
        
        Returns:
            The key char(s).
        """
        return self.get("keys.rewind_back")
    @key_rewind_back.setter
    def key_rewind_back(self, value: str): self.set("keys.rewind_back", value)
    
    @property
    def key_volume_up(self) -> str:
        """The volume up key.
        
        Returns:
            The key char(s).
        """
        return self.get("keys.volume_up")
    @key_volume_up.setter
    def key_volume_up(self, value: str): self.set("keys.volume_up", value)
    
    @property
    def key_volume_down(self) -> str:
        """The volume down key.
        
        Returns:
            The key char(s).
        """
        return self.get("keys.volume_down")
    @key_volume_down.setter
    def key_volume_down(self, value: str): self.set("keys.volume_down", value)
    
    # ! Debag
    @property
    def logging(self) -> bool:
        """Enabling and disabling logging.
        
        Returns:
            On or off.
        """
        return self.get("debag.logging")
    @logging.setter
    def logging(self, value: bool): self.set("debag.logging", value)



import os
import glob
import asyncio
# > Graphics
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding, _Bindings
from textual.containers import Horizontal, Vertical, Container
from textual.widgets import Header, Footer, Static, Label, Button, Input, ListView
# > Image Works
from PIL import Image
# > Typing
from typing import Optional, Literal, Tuple, List, Type, Union
# > Local Imports
from .config import SeaPlayerConfig
from .types import Cacher, Environment
from .codeсbase import CodecBase
from .languages import LanguageLoader
from .screens import Unknown, Configurate, UNKNOWN_OPEN_KEY
from .codecs import codecs
from .functions import (
    aiter, awrap,
    image_from_bytes,
    get_sound_basename,
    aio_check_status_code
)
from .objects import (
    Nofy,
    LogMenu,
    CallNofy,
    PlayListView,
    AsyncImageLabel,
    StandartImageLabel,
    IndeterminateProgress
)
from .units import (
    __title__,
    __version__,
    __author__,
    __email__,
    __url__,
    CSS_LOCALDIR,
    CONFIG_FILEPATH,
    ASSETS_DIRPATH,
    IMGPATH_IMAGE_NOT_FOUND,
    RESAMPLING_SAFE,
    LOCALDIR,
    ENABLE_PLUGIN_SYSTEM,
    CACHE_DIRPATH,
    LANGUAGES_DIRPATH
)
# > Plugin System Init
if ENABLE_PLUGIN_SYSTEM:
    from .plug import PluginLoader

# ! Main Functions
def build_bindings(config: SeaPlayerConfig, ll: LanguageLoader):
    yield Binding(config.key_quit, "quit", ll.get("footer.quit"))
    yield Binding("c,с", "push_screen('configurate')", ll.get("footer.configurate"))
    if config.logging:
        yield Binding("l,д", "app.toggle_class('LogMenu', '--hidden')", ll.get("footer.logs"))
    yield Binding(config.key_rewind_back, "minus_rewind", ll.get('footer.rewind.minus').format(sec=config.rewind_count_seconds))
    yield Binding(config.key_rewind_forward, "plus_rewind", ll.get('footer.rewind.plus').format(sec=config.rewind_count_seconds))
    yield Binding(config.key_volume_down, "minus_volume", ll.get('footer.volume.minus').format(per=round(config.volume_change_percent*100)))
    yield Binding(config.key_volume_up, "plus_volume", ll.get('footer.volume.plus').format(per=round(config.volume_change_percent*100)))
    yield Binding("ctrl+s", "screenshot", ll.get('footer.screenshot'))
    yield Binding(UNKNOWN_OPEN_KEY, "push_screen('unknown')", show=False)

# ! Main
class SeaPlayer(App):
    # ! Textual Configuration
    TITLE = f"{__title__} v{__version__}"
    CSS_PATH = [
        os.path.join(CSS_LOCALDIR, "seaplayer.tcss"),
        os.path.join(CSS_LOCALDIR, "configurate.tcss"),
        os.path.join(CSS_LOCALDIR, "unknown.tcss"),
        os.path.join(CSS_LOCALDIR, "objects.tcss")
    ]
    SCREENS = {
        "unknown": Unknown(id="screen_unknown"),
        "configurate": Configurate(id="screen_configurate")
    }
    ENABLE_COMMAND_PALETTE = False
    
    # ! SeaPlayer Configuration
    cache: Cacher = Cacher(CACHE_DIRPATH)
    """An image of a class for caching variables."""
    config: SeaPlayerConfig = SeaPlayerConfig(CONFIG_FILEPATH)
    """The image of the SeaPlayer configuration file."""
    ll: LanguageLoader = LanguageLoader(LANGUAGES_DIRPATH, config.lang)
    """An image of the class for receiving the loaded SeaPlayer translation. With the translation uploaded from the `seaplayer/langs/` directory."""
    image_type: Optional[Union[Type[AsyncImageLabel], Type[StandartImageLabel]]] = None
    env = Environment(
        {
            "seaplayer": {
                "codecs": [*codecs],
                "codecs_kwargs": {
                    "sound_fonts_path": config.sound_font_path,
                    "sound_device_id": config.output_sound_device_id
                }
            }
        }
    )
    
    # ! Bindings
    BINDINGS = list(build_bindings(config, ll))
    
    # ! Template Configuration
    currect_sound: Optional[CodecBase] = None
    currect_sound_index: Optional[int] = None
    currect_volume: float = cache.var("currect_volume", 1.0)
    """The current volume value (cached)."""
    last_playback_status: Optional[Literal[0, 1, 2]] = None
    playback_mode: int = cache.var("playback_mode", 0)
    """The current playback mode (cached)."""
    block_select: bool = False
    last_handlered_values: List[str] = []
    started: bool = True
    
    # ! Init Objects
    log_menu = LogMenu(
        enable_logging=config.logging,
        wrap=True, highlight=True, markup=True
    )
    
    # ! Log Functions
    info = log_menu.info
    error = log_menu.error
    warn = log_menu.warn
    exception = log_menu.exception
    
    # ! App Init
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if ENABLE_PLUGIN_SYSTEM:
            self.plugin_loader = PluginLoader(self)
            self.plugin_loader.on_init()
            self.update_bindings()
    
    def update_bindings(self) -> None:
        bindings = list(build_bindings(self.config, self.ll))
        if ENABLE_PLUGIN_SYSTEM:
            for _binding in self.plugin_loader.on_bindings():
                if _binding is not None:
                    bindings.append(_binding)
        self._bindings = _Bindings(bindings)
    
    # ! Inherited Functions
    async def action_push_screen(self, screen: str) -> None:
        if self.SCREENS[screen].id != self.screen.id:
            await super().action_push_screen(screen)
    
    # ! Nofy's
    def nofy(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> None:
        """Creating a temporary notification.
        
        Args:
            text (str): The text of the notification.
            life_time (float, optional): The time in seconds after which the notification will disappear. Defaults to 3.
            dosk (Literal[\"bottom\", \"left\", \"right\", \"top\"], optional): Regarding the screen. Defaults to "top".
        """
        self.screen.mount(Nofy(text, life_time, dosk))
    
    async def aio_nofy(
        self,
        text: str,
        life_time: float=3,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> None:
        """Creating a temporary notification.
        
        Args:
            text (str): The text of the notification.
            life_time (float, optional): The time in seconds after which the notification will disappear. Defaults to 3.
            dosk (Literal[\"bottom\", \"left\", \"right\", \"top\"], optional): Regarding the screen. Defaults to "top".
        """
        await self.screen.mount(Nofy(text, life_time, dosk))
    
    def callnofy(
        self,
        text: str,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> CallNofy:
        """Creating a notification.
        
        Args:
            text (str): The text of the notification.
            dosk (Literal[\"bottom\", \"left\", \"right\", \"top\"], optional): Regarding the screen. Defaults to "top".
        
        Returns:
            CallNofy: To delete the notification image, use the `CallNofy.remove()` method.
        """
        cn = CallNofy(text, dosk)
        self.screen.mount(cn)
        self.install_screen
        return cn
    
    async def aio_callnofy(
        self,
        text: str,
        dosk: Literal["bottom", "left", "right", "top"]="top"
    ) -> CallNofy:
        """Creating a notification.
        
        Args:
            text (str): The text of the notification.
            dosk (Literal[\"bottom\", \"left\", \"right\", \"top\"], optional): Regarding the screen. Defaults to "top".
        
        Returns:
            CallNofy: To delete the notification image, use the `CallNofy.remove()` method.
        """
        cn = CallNofy(text, dosk)
        await self.screen.mount(cn)
        return cn
    
    # ! Get Current Sound
    def gcs(self) -> Optional[CodecBase]:
        """Getting the currently selected sound.
        
        Returns:
            Optional[CodecBase]: The image of the codec in which the sound is wrapped.
        """
        return self.currect_sound
    
    async def aio_gcs(self) -> Optional[CodecBase]:
        """Getting the currently selected sound.
        
        Returns:
            Optional[CodecBase]: The image of the codec in which the sound is wrapped.
        """
        return self.currect_sound
    
    # ! Get Current Sound Status Text
    def get_sound_tstatus(self, sound: CodecBase) -> str:
        """Getting the audio status in text format in the language selected by the user.
        
        Args:
            sound (CodecBase): The image of the codeс in which the sound is wrapped.
        
        Returns:
            str: Audio status in text format in the language selected by the user.
        """
        if sound.playing:
            if sound.paused:
                return self.ll.get("sound.status.paused")
            else:
                return self.ll.get("sound.status.playing")
        return self.ll.get("sound.status.stopped")
    
    def get_sound_selected_label_text(self, sound: Optional[CodecBase]=None) -> str:
        """Generating a string for `self.music_selected_label`.
        
        Args:
            sound (Optional[CodecBase], optional): The image of the codeс in which the sound is wrapped. Defaults to None.
        
        Returns:
            str: String for `self.music_selected_label`.
        """
        if sound is None:
            sound = self.gcs()
        if sound is not None:
            return f"({self.get_sound_tstatus(sound)}): {get_sound_basename(sound)}"
        return self.ll.get("player.bar.sound.none")
    
    async def aio_get_sound_selected_label_text(self, sound: Optional[CodecBase]=None) -> str:
        """Generating a string for `self.music_selected_label`.
        
        Args:
            sound (Optional[CodecBase], optional): The image of the codeс in which the sound is wrapped. Defaults to None.
        
        Returns:
            str: String for `self.music_selected_label`.
        """
        if sound is None:
            sound = await self.aio_gcs()
        if sound is not None:
            return f"({self.get_sound_tstatus(sound)}): {get_sound_basename(sound)}"
        return self.ll.get("player.bar.sound.none")
    
    # ! Update Selected Label Text
    def update_select_label(self, sound: Optional[CodecBase]=None) -> None:
        """Updating the string in `self.music_selected_label`.
        
        Args:
            sound (Optional[CodecBase], optional): The image of the codeс in which the sound is wrapped. Defaults to None.
        """
        self.player_selected_label.update(self.get_sound_selected_label_text(sound))
    
    async def aio_update_select_label(self, sound: Optional[CodecBase]=None) -> None:
        """Updating the string in `self.music_selected_label`.
        
        Args:
            sound (Optional[CodecBase], optional): The image of the codeс in which the sound is wrapped. Defaults to None.
        """
        self.player_selected_label.update(await self.aio_get_sound_selected_label_text(sound))
    
    # ! Update Selected Image
    async def aio_update_select_image(self, sound: Optional[CodecBase]) -> None:
        image = None
        if sound is not None:
            if sound.icon_data is not None:
                image = image_from_bytes(sound.icon_data)
        await self.player_image.update_image(image)
    
    # ! Update Currect Sound
    async def aio_update_currect_sound(self) -> None:
        self.currect_sound = self.playlist_view.currect_sound
        self.currect_sound_index = self.playlist_view.currect_sound_index
        if self.currect_sound is not None:
            self.currect_sound.set_volume(self.currect_volume)
    
    # ! Switch Mode Button
    def gpms(
        self,
        modes: Tuple[str, str, str]=(
            ll.get("player.button.mode.play"),
            ll.get("player.button.mode.replay_sound"),
            ll.get("player.button.mode.replay_list")
        )
    ) -> str:
        return modes[self.playback_mode]
    
    def switch_playback_mode(self) -> None:
        if self.playback_mode == 2:
            self.playback_mode = 0
        else:
            self.playback_mode += 1
    
    # ! Callback Functions
    async def get_sound_seek(self) -> Tuple[str, Optional[float], Optional[float]]:
        if (sound:=await self.aio_gcs()) is not None:
            pos = sound.get_pos()
            minutes, seconds = round(pos // 60), round(pos % 60)
            return f"{minutes}:{str(seconds).rjust(2,'0')} | {str(round(sound.get_volume()*100)).rjust(3)}%", pos, sound.duration
        return "0:00 |   0%", None, None
    
    # ! Loop Functions
    async def playback_control_loop(self) -> None:
        while self.started:
            sound = await self.aio_gcs()
            if sound is not None:
                if self.last_playback_status is not None:
                    if self.playback_mode != 0:
                        status = await aio_check_status_code(sound)
                        if (self.last_playback_status == 1) and (status == 0):
                            if self.playback_mode == 1:
                                self.currect_sound.play()
                                await self.aio_update_select_label(sound)
                                self.last_playback_status = 0
                            elif self.playback_mode == 2:
                                self.block_select = True
                                await self.playlist_view.aio_select_next_sound()
                                await self.aio_update_currect_sound()
                                if self.currect_sound is not None:
                                    self.currect_sound.play()
                                await self.aio_update_select_label(self.currect_sound)
                                await self.aio_update_select_image(self.currect_sound)
                                self.last_playback_status = 0
                                self.block_select = False
            await asyncio.sleep(0.1)
    
    # ! Mounting Function
    def compose(self) -> ComposeResult:
        # * Other
        self.info("---")
        self.info(f"{__title__} [#60fdff]v{__version__}[/#60fdff] from {__author__} ({__email__})")
        self.info(f"Source          : {__url__}")
        self.info(f"Environment     : {repr(self.env)}")
        self.info(f"Codecs          : {repr(self.env['seaplayer']['codecs'])}")
        self.info(f"Config Path     : {repr(self.config.filepath)}")
        self.info(f"CSS Dirpath     : {repr(CSS_LOCALDIR)}")
        self.info(f"Assets Dirpath  : {repr(ASSETS_DIRPATH)}")
        self.info(f"Codecs Kwargs   : {repr(self.env['seaplayer']['codecs_kwargs'])}")
        self.info(f"Language Loader : {repr(self.ll)}")
        
        # * Play Screen
        self.player_box = Container(classes="player-box")
        self.player_box.border_title = self.ll.get("player")
        
        # * Image Object Init
        self.player_selected_label = Label(
            self.get_sound_selected_label_text(),
            classes="player-selected-label"
        )
        if self.config.image_update_method == "sync":
            self.image_type = StandartImageLabel
        elif self.config.image_update_method == "async":
            self.image_type = AsyncImageLabel
        else:
            raise RuntimeError("The configuration 'image_update_method' is incorrect.")
        
        self.player_image = self.image_type(
            Image.open(IMGPATH_IMAGE_NOT_FOUND),
            resample=RESAMPLING_SAFE[self.config.image_resample_method]
        )
        self.info(f"The picture from the media file is rendered using the {repr(self.config.image_update_method)} method.")
        
        # * Compositions Screen
        self.playlist_box = Container(classes="playlist-box")
        self.playlist_box.border_title = self.ll.get("playlist")
        self.playlist_view = PlayListView(classes="playlist-view")
        
        self.playlist_add_sound_input = Input(
            classes="playlist-add-sound-input", id="addsoundinput",
            placeholder=self.ll.get("playlist.input.placeholder")
        )
        self.player_button_mode_switch = Button(
            self.gpms(),
            id="switch-playback-mode",
            variant="primary",
            classes="button-sound-control"
        )
        
        # * Adding
        yield Header()
        with self.player_box:
            with Vertical():
                with Container(classes="player-visual-panel"):
                    yield self.player_image
                with Container(classes="player-contol-panel"):
                    yield self.player_selected_label
                    yield IndeterminateProgress(getfunc=self.get_sound_seek)
                    with Horizontal(classes="box-buttons-sound-control"):
                        yield Button(
                            self.ll.get("player.button.pause"),
                            id="button-pause",
                            variant="success",
                            classes="button-sound-control"
                        )
                        yield Static(classes="pass-one-width")
                        yield Button(
                            self.ll.get("player.button.psu"),
                            id="button-play-stop",
                            variant="warning",
                            classes="button-sound-control"
                        )
                        yield Static(classes="pass-one-width")
                        yield self.player_button_mode_switch
        with self.playlist_box:
            yield self.playlist_view
            yield self.playlist_add_sound_input
        yield self.log_menu
        yield Footer()
        self.pcw = self.run_worker(
            self.playback_control_loop,
            name="Playback Control Loop",
            group="seaplayer-main",
            description="Control of playback modes and status updates.",
            thread=True
        )
        self.info("---")
    
    # ! Currect Sound Controls
    async def currect_sound_stop(self, sound: Optional[CodecBase]=None):
        """Stops playback of the currently selected sound.
        
        Args:
            sound (Optional[CodecBase], optional): Сurrently selected sound. Defaults to None.
        """
        if sound is None:
            sound = await self.aio_gcs()
        if sound is not None:
            self.last_playback_status = 0
            sound.stop()
    
    async def currect_sound_play(self, sound: Optional[CodecBase]=None):
        """Plays playback of the currently selected sound.
        
        Args:
            sound (Optional[CodecBase], optional): Сurrently selected sound. Defaults to None.
        """
        if sound is None:
            sound = await self.aio_gcs()
        if sound is not None:
            self.last_playback_status = 1
            sound.play()
    
    async def currect_sound_pause(self, sound: Optional[CodecBase]=None):
        """Pauses the currently selected sound.
        
        Args:
            sound (Optional[CodecBase], optional): Сurrently selected sound. Defaults to None.
        """
        if sound is None:
            sound = await self.aio_gcs()
        if sound is not None:
            self.last_playback_status = 3
            sound.pause()
    
    async def currect_sound_unpause(self, sound: Optional[CodecBase]=None):
        """Unpauses the currently selected sound.
        
        Args:
            sound (Optional[CodecBase], optional): Сurrently selected sound. Defaults to None.
        """
        if sound is None:
            sound = await self.aio_gcs()
        if sound is not None:
            self.last_playback_status = 1
            sound.unpause()
    
    # ! Sound Controls
    async def adding_sounds_loader(self, handlered_values: List[str]) -> None:
        added_oks = 0
        loading_nofy = await self.aio_callnofy(
            self.ll.get("nofys.sound.found").format(count=len(handlered_values))
        )
        self.env['seaplayer']['codecs'].sort(key=lambda x: x.codec_priority)
        codec: CodecBase
        async for value in aiter(handlered_values):
            sound = None
            async for codec in aiter(self.env['seaplayer']['codecs']):
                self.info(f"Attempt to load via {repr(codec)}")
                try:
                    if hasattr(codec, "aio_is_this_codec"):
                        this_codec = await codec.aio_is_this_codec(value)
                    else:
                        this_codec = codec.is_this_codec(value)
                    if this_codec:
                        if hasattr(codec, "__aio_init__"):
                            try:
                                sound: CodecBase = await codec.__aio_init__(value, **self.env['seaplayer']['codecs_kwargs'])
                            except Exception as e:
                                sound = None
                                raise e
                        else:
                            try:
                                sound: CodecBase = codec(value, **self.env['seaplayer']['codecs_kwargs'])
                            except Exception as e:
                                sound = None
                                raise e
                        if sound is not None:
                            if not await self.playlist_view.aio_exist_sound(sound):
                                await self.playlist_view.aio_add_sound(sound)
                                self.info(f"Song added: {repr(sound)}")
                                added_oks += 1
                                break
                except FileNotFoundError:
                    self.error(f"The file does not exist or is a directory: {repr(value)}")
                    break
                except OSError:
                    pass
                except Exception as e:
                    self.exception(e)
            if sound is None:
                self.error(f"The sound could not be loaded: {repr(value)}")
        await loading_nofy.remove()
        self.info(f"Added [cyan]{added_oks}[/cyan] songs!")
        await self.aio_nofy(self.ll.get("nofys.sound.added").format(count=added_oks))
    
    # ! Worker Functions
    async def pl_select(self) -> None:
        if self.playlist_view.currect_sound_index != self.currect_sound_index:
            if self.currect_sound is not None:
                self.currect_sound.stop()
            await self.aio_update_currect_sound()
            await self.aio_update_select_label(self.currect_sound)
            await self.aio_update_select_image(self.currect_sound)
    
    # ! Playlist Actions
    @on(ListView.Selected, ".playlist-view")
    async def pl_select_worker(self) -> None:
        if not self.block_select:
            self.run_worker(
                self.pl_select,
                name="Playlist Select",
                group="seaplayer-temp",
                description="Update: status, images and so on."
            )
    
    # ! Input Actions
    @on(Input.Submitted, "#addsoundinput")
    async def input_add_sound(self) -> None:
        value = self.playlist_add_sound_input.value
        self.playlist_add_sound_input.value = ""
        if len(value.replace(" ", "")) > 0:
            self.run_worker(
                awrap(self.adding_sounds_handler, value),
                name="Adding Sounds Handler",
                group="seaplayer-temp"
            )
    
    # ! Button Actions
    @on(Button.Pressed, "#switch-playback-mode")
    async def on_button_pressed(self) -> None:
        self.switch_playback_mode()
        self.player_button_mode_switch.label = self.gpms()
    
    @on(Button.Pressed, "#button-pause")
    async def bp_pause_unpause(self) -> None:
        if (sound:=await self.aio_gcs()) is not None:
            if sound.playing:
                await self.currect_sound_pause(sound)
            await self.aio_update_select_label(sound)
    
    @on(Button.Pressed, "#button-play-stop")
    async def bp_play_stop(self) -> None:
        if (sound:=await self.aio_gcs()) is not None:
            if sound.playing:
                if sound.paused:
                    await self.currect_sound_unpause(sound)
                else:
                    await self.currect_sound_stop(sound)
            else:
                await self.currect_sound_play(sound)
            await self.aio_update_select_label(sound)
    
    # ! Input Submits
    async def adding_sounds_handler(self, value: str) -> None:
        handlered_values = []
        try: handlered_values = glob.glob(value, recursive=self.config.recursive_search)
        except: pass
        if ENABLE_PLUGIN_SYSTEM:
            for vhr in self.plugin_loader.value_handlers:
                handlered_values += vhr(value)
        self.info(f"Submit 'plus_sound' values: {repr(handlered_values)}")
        if len(handlered_values) > 0:
            self.run_worker(
                awrap(self.adding_sounds_loader, handlered_values),
                name="Adding Sounds Loader",
                group="seaplayer-temp",
                description="The process of adding sounds to a playlist."
            )
    
    # ! Keys Actions
    async def action_plus_rewind(self):
        if (sound:=await self.aio_gcs()) is not None:
            sound.set_pos(sound.get_pos()+self.config.rewind_count_seconds)
    
    async def action_minus_rewind(self):
        if (sound:=await self.aio_gcs()) is not None:
            sound.set_pos(sound.get_pos()-self.config.rewind_count_seconds)
    
    async def action_plus_volume(self) -> None:
        if (sound:=await self.aio_gcs()) is not None:
            if (vol:=round(sound.get_volume()+self.config.volume_change_percent, 2)) <= self.config.max_volume_percent:
                self.currect_volume = vol
                sound.set_volume(vol)
    
    async def action_minus_volume(self) -> None:
        if (sound:=await self.aio_gcs()) is not None:
            if (vol:=round(sound.get_volume()-self.config.volume_change_percent, 2)) >= 0:
                self.currect_volume = vol
                sound.set_volume(vol)
    
    async def action_screenshot(self) -> None:
        path = self.save_screenshot(path=LOCALDIR)
        self.info(f"Screenshot saved to: {repr(path)}")
        await self.aio_nofy(self.ll.get("nofys.screenshot.saved").format(path=repr(path)))
    
    async def action_quit(self) -> None:
        """The function called by our when the SeaPlayer stops working."""
        self.started = False
        if ENABLE_PLUGIN_SYSTEM:
            await self.plugin_loader.on_quit()
        if (sound:=await self.aio_gcs()) is not None:
            sound.unpause()
            sound.stop()
        return await super().action_quit()
    
    # ! On Functions
    async def on_compose(self) -> None:
        if ENABLE_PLUGIN_SYSTEM:
            self.run_worker(
                self.plugin_loader.on_compose,
                name="On Compose",
                group="pluginloader",
                description="<method PluginLoader.on_compose>"
            )
    
    def on_ready(self, *args, **kwargs) -> None:
        """A function called when the SeaPlayer is completely confused."""
        if ENABLE_PLUGIN_SYSTEM:
            self.run_worker(
                self.plugin_loader.on_ready,
                name="On Ready",
                group="pluginloader",
                description="<method PluginLoader.on_ready>"
            )
    
    # ! Other
    def run(self, *args, **kwargs):
        if ENABLE_PLUGIN_SYSTEM:
            self.plugin_loader.on_run()
        super().run(*args, **kwargs)



import os
import asyncio
import aiofiles
import subprocess
from tempfile import mkstemp
# > IO's
from urlopen2 import URLFile, AsyncURLFile
# > Sound Works
from playsoundsimple import Sound
from playsoundsimple.sound import FPType, getfp
from playsoundsimple.units import DEFAULT_SOUND_FONTS_PATH
from playsoundsimple.exceptions import FileTypeError
# > Typing
from typing import Optional

# ! Vars
FLUID_SYNTH_PATH = 'fluidsynth'

# ! Main Class
class AnySound(Sound):
    @staticmethod
    async def aio_from_midi(
        fp: FPType,
        sound_fonts_path: Optional[str]=None,
        **kwargs
    ):
        path, is_temp = getfp(fp, filetype=".midi")
        sound_fonts_path = sound_fonts_path or DEFAULT_SOUND_FONTS_PATH
        if path is None:
            raise FileTypeError(fp)
        npath = mkstemp(suffix=".wav")[1]
        
        process = await asyncio.create_subprocess_exec(
            FLUID_SYNTH_PATH, "-ni", sound_fonts_path, path, "-F", npath, "-q",
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        await process.wait()
        if is_temp:
            try:
                os.remove(path)
            except:
                pass
        return AnySound(npath, **{"is_temp": True, **kwargs})
    
    @staticmethod
    def from_midi(
        fp: FPType,
        sound_fonts_path: Optional[str]=None,
        **kwargs
    ):
        path, is_temp = getfp(fp, filetype=".midi")
        sound_fonts_path = sound_fonts_path or DEFAULT_SOUND_FONTS_PATH
        if path is None:
            raise FileTypeError(fp)
        npath = mkstemp(suffix=".wav")[1]
        subprocess.call(
            [FLUID_SYNTH_PATH, "-ni", sound_fonts_path, path, "-F", npath, "-q"],
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        if is_temp:
            try:
                os.remove(path)
            except:
                pass
        return AnySound(npath, **{"is_temp": True, **kwargs})
    
    @staticmethod
    def from_url(url: str, **kwargs):
        path = mkstemp(suffix=".bin")[1]
        with open(path, "wb") as tempfile:
            with URLFile(url, tempfile) as urlfile:
                urlfile.fulling()
        return AnySound(path, is_temp=True, **kwargs)
    
    @staticmethod
    async def aio_from_url(url: str, **kwargs):
        path = mkstemp(suffix=".bin")[1]
        async with aiofiles.open(path, "wb") as tempfile:
            async with AsyncURLFile(url, tempfile) as urlfile:
                await urlfile.fulling()
        return AnySound(path, is_temp=True, **kwargs)



from typing import TypeVar, Generic, Optional, Dict

# ! Types
T = TypeVar('T')
K = TypeVar('K')
G = TypeVar('G')

# ! Main Class
class Environment(Generic[G, K, T]):
    def __init__(
        self,
        setup_data: Dict[G, Dict[K, T]]={}
    ) -> None:
        self.env: Dict[G, Dict[K, T]] = setup_data
    
    # ? Magic Methods
    def __str__(self) -> str:
        return str(self.env)
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.env)})"
    
    def __getitem__(self, group: K) -> Dict[K, T]:
        if group not in list(self.env.keys()):
            self.env[group] = {}
        return self.env[group]
    
    def __setitem__(self, group: K, value: Dict[K, T]) -> None:
        assert isinstance(value, dict)
        self.env[group] = value
    
    # ? Main Methods
    def set(self, group: G, key: K, value: T) -> None:
        if group not in list(self.env.keys()):
            self.env[group] = {}
        self.env[group][key] = value
    
    def get(self, group: G, key: K, default: Optional[T]=None) -> Optional[T]:
        if group in list(self.env.keys()):
            return self.env[group].get(key, default)



# ! Metadata
__prog_name__ = 'seaplug'
__prog_title__ = 'SeaPlayer Plugins System CLI'
__prog_version__ = '0.3.0'
__prog_author__ = 'Romanin'
__prog_email__ = 'semina054@gmail.com'

# ! Vars
CREATE_DEFAULT_CODE = """\
from seaplayer.plug import PluginBase

class Plugin(PluginBase):
    pass

__plugin__ = Plugin
"""


from textual.widgets import Label, ListItem, ListView
# > Typing
from typing import Optional, Dict, Sequence
# > Local Import's
from .Labels import FillLabel
from ..codeсbase import CodecBase
from ..functions import get_sound_basename, aiter

# ! Children Classes
class PlayListViewItem(ListItem):
    DEFAULT_CSS = """
    PlayListViewItem {
        height: 4;
    }
    PlayListViewItem .title-label {
        height: 1;
        color: #cacaca;
    }
    PlayListViewItem .subtitle-label {
        height: 1;
        color: #a9a9a9;
    }
    PlayListViewItem FillLabel {
        height: 1;
    }
    """
    
    def __init__(
        self,
        sound: CodecBase,
        sound_sha1: str,
        title: str="",
        first_subtitle: str="",
        second_subtitle: str=""
    ) -> None:
        super().__init__()
        self.title_label = Label(title, classes="title-label")
        self.first_subtitle_label = Label(f" {first_subtitle}", classes="subtitle-label")
        self.second_subtitle_label = Label(f" {second_subtitle}", classes="subtitle-label")
        self.sound = sound
        self.sound_sha1 = sound_sha1
        
        self.compose_add_child(self.title_label)
        self.compose_add_child(self.first_subtitle_label)
        self.compose_add_child(self.second_subtitle_label)
        self.compose_add_child(FillLabel('─'))
    
    async def update_labels(
        self,
        title: Optional[str]=None,
        first_subtitle: Optional[str]=None,
        second_subtitle: Optional[str]=None,
    ) -> None:
        if title is not None: self.title_label.update(title)
        if first_subtitle is not None: self.first_subtitle_label.update(first_subtitle)
        if second_subtitle is not None: self.second_subtitle_label.update(second_subtitle)

# ! Main Class
class PlayListView(ListView):
    DEFAULT_CSS = """
    PlayListView {
        height: 1fr;
    }
    """
    highlighted_child: Optional[PlayListViewItem]
    children: Sequence[PlayListViewItem]
    
    @property
    def currect_sound(self) -> Optional[CodecBase]:
        if self.highlighted_child is not None:
            return self.highlighted_child.sound
    
    @property
    def currect_sound_index(self) -> Optional[int]:
        return self.index
    
    @property
    def currect_sound_sha1(self) -> Optional[str]:
        if self.highlighted_child is not None:
            return self.highlighted_child.sound_sha1
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.sounds: Dict[str, CodecBase] = {}
    
    # ! Sync Methods
    def add_sound(self, sound: CodecBase) -> str:
        sound_sha1 = sound.__sha1__(65536)
        self.sounds[sound_sha1] = sound
        self.append(
            PlayListViewItem(
                sound,
                sound_sha1,
                f"{get_sound_basename(sound)}",
                "{duration} sec, {channel_mode}, {samplerate} Hz, {bitrate} kbps".format(
                    duration=round(sound.duration),
                    channel_mode="Mono" if sound.channels <= 1 else "Stereo",
                    samplerate=round(sound.samplerate),
                    bitrate=round(sound.bitrate / 1000)
                ),
                sound.__namerepr__()
            )
        )
        return sound_sha1
    
    def exist_sound(self, sound: CodecBase) -> bool:
        check_sound_sha1 = sound.__sha1__(65536)
        child: PlayListViewItem
        for child in self.children:
            if child.sound_sha1 == check_sound_sha1:
                return True
        return False
    
    def get_sound_by_index(self, index: int) -> CodecBase:
        if len(self.children) > index:
            return self.children[index].sound
        raise IndexError(f"There is no `Sound` with such an index: {index}.")
    
    def get_sound_by_sha1(self, sha1: str) -> CodecBase:
        for child in self.children:
            if child.sound_sha1 == sha1:
                return child.sound
        raise IndexError(f"There is no `Sound` with such a sha1: {sha1}")
    
    def get_child_by_index(self, index: int) -> PlayListViewItem:
        if len(self.children) > index:
            return self.children[index]
        raise IndexError(f"There is no `PlayListViewItem` with such an index: {repr(index)}.")
    
    def get_child_by_sha1(self, sha1: str) -> PlayListViewItem:
        for child in self.children:
            if child.sound_sha1 == sha1:
                return child
        raise IndexError(f"There is no `PlayListViewItem` with such a sha1: {repr(sha1)}.")
    
    def get_next_sound_index(self) -> Optional[int]:
        if self.currect_sound_index is not None:
            if len(self.children) > (self.currect_sound_index + 1):
                return self.currect_sound_index + 1
            else:
                return 0
    
    def select_child(self, widget: PlayListViewItem) -> None:
        self._on_list_item__child_clicked(ListItem._ChildClicked(widget))
    
    def select_by_index(self, index: int) -> None:
        self.select_child(self.get_child_by_index(index))
    
    def select_by_sha1(self, sha1: str) -> None:
        self.select_child(self.get_child_by_index(sha1))
    
    def select_next_sound(self) -> None:
        index = self.get_next_sound_index()
        if index is not None:
            self.select_by_index(index)
    
    # ! Async Methods
    async def aio_exist_sound(self, sound: CodecBase) -> bool:
        check_sound_sha1 = await sound.__aio_sha1__(65536)
        child: PlayListViewItem
        async for child in aiter(self.children):
            if child.sound_sha1 == check_sound_sha1:
                return True
        return False
    
    async def aio_add_sound(self, sound: CodecBase) -> str:
        sound_sha1 = await sound.__aio_sha1__(65536)
        self.sounds[sound_sha1] = sound
        await self.append(
            PlayListViewItem(
                sound, sound_sha1,
                f"{get_sound_basename(sound)}",
                "{duration} sec, {channel_mode}, {samplerate} Hz, {bitrate} kbps".format(
                    duration=round(sound.duration),
                    channel_mode="Mono" if sound.channels <= 1 else "Stereo",
                    samplerate=round(sound.samplerate),
                    bitrate=round(sound.bitrate / 1000)
                ),
                sound.__namerepr__()
            )
        )
        return sound_sha1
    
    async def aio_get_sound_by_index(self, index: int) -> CodecBase:
        if len(self.children) > index:
            return self.children[index].sound
        raise IndexError(f"There is no `Sound` with such an index: {index}.")
    
    async def aio_get_sound_by_sha1(self, sha1: str) -> CodecBase:
        async for child in aiter(self.children):
            if child.sound_sha1 == sha1:
                return child.sound
        raise IndexError(f"There is no `Sound` with such a sha1: {sha1}")
    
    async def aio_get_child_by_index(self, index: int) -> PlayListViewItem:
        if len(self.children) > index:
            return self.children[index]
        raise IndexError(f"There is no `PlayListViewItem` with such an index: {repr(index)}.")
    
    async def aio_get_child_by_sha1(self, sha1: str) -> PlayListViewItem:
        async for child in aiter(self.children):
            if child.sound_sha1 == sha1:
                return child
        raise IndexError(f"There is no `PlayListViewItem` with such a sha1: {repr(sha1)}.")
    
    async def aio_get_next_sound_index(self) -> Optional[int]:
        if self.currect_sound_index is not None:
            if len(self.children) > (self.currect_sound_index + 1):
                return self.currect_sound_index + 1
            else:
                return 0
    
    async def aio_select_child(self, widget: PlayListViewItem) -> None:
        self._on_list_item__child_clicked(ListItem._ChildClicked(widget))
    
    async def aio_select_by_index(self, index: int) -> None:
        await self.aio_select_child(await self.aio_get_child_by_index(index))
    
    async def aio_select_by_sha1(self, sha1: str) -> None:
        await self.aio_select_child(await self.aio_get_child_by_index(sha1))
    
    async def aio_select_next_sound(self) -> None:
        index = await self.aio_get_next_sound_index()
        if index is not None:
            await self.aio_select_by_index(index)